from math import modf
from typing import Tuple


def split_seconds(seconds: float) -> Tuple[int, int]:
    """Split seconds into whole seconds and nanoseconds."""
    fraction, whole = modf(seconds)
    return int(whole), int(fraction * 1_000_000_000.0)


def unsplit_seconds(seconds: int, nanos: int) -> float:
    """Merge whole seconds and nanoseconds back to normal seconds."""
    return float(seconds) + float(nanos) / 1_000_000_000.0



from abc import abstractmethod
from typing import Optional, Protocol

from pure_protobuf.interfaces._repr import Repr
from pure_protobuf.interfaces._vars import FieldT


class Merge(Repr, Protocol[FieldT]):
    """Defines how to merge two values of the same field."""

    @abstractmethod
    def __call__(
        self,
        __accumulator: Optional[FieldT],
        __value: Optional[FieldT],
    ) -> Optional[FieldT]:
        raise NotImplementedError(f"{type(self)}.__call__()")



from typing import Generic, List, Optional

from pure_protobuf._accumulators import AccumulateMessages
from pure_protobuf.interfaces._repr import ReprWithInner
from pure_protobuf.interfaces._vars import FieldT, MessageT, RecordT
from pure_protobuf.interfaces.merge import Merge


class MergeLastOneWins(Merge[FieldT], Generic[FieldT]):
    """
    Always returns the right-hand side.

    See Also:
        - https://developers.google.com/protocol-buffers/docs/encoding#last-one-wins
    """

    def __call__(self, _lhs: Optional[FieldT], rhs: Optional[FieldT]) -> Optional[FieldT]:
        return rhs


class MergeConcatenate(Merge[List[RecordT]], Generic[RecordT]):
    def __call__(
        self,
        lhs: Optional[List[RecordT]],
        rhs: Optional[List[RecordT]],
    ) -> Optional[List[RecordT]]:
        if lhs is None:
            return rhs
        if rhs is None:
            return lhs
        lhs.extend(rhs)
        return lhs


class MergeMessages(Merge[MessageT], ReprWithInner):
    __slots__ = ("inner",)

    # noinspection PyProtocol
    def __init__(self, inner: AccumulateMessages[MessageT]) -> None:
        self.inner = inner

    def __call__(self, lhs: Optional[MessageT], rhs: Optional[MessageT]) -> Optional[MessageT]:
        """
        Merge the two messages.

        Returns:
            Merged message: it'll be either possibly updated `lhs` or non-updated `rhs`.

        Notes:
            - **Never** reuse either `lhs` or `rhs` after calling this method, consider
            them consumed.
        """
        if rhs is not None:
            # noinspection PyTypeChecker
            return self.inner(lhs, (rhs,))
        return lhs



from typing import Callable, Generic, Iterator, TypeVar

from typing_extensions import ParamSpec

from pure_protobuf.interfaces._repr import ReprWithInner

P = ParamSpec("P")
R = TypeVar("R")


class ReadCallback(ReprWithInner, Generic[P, R]):
    """Convert a reader, which returns a singular scalar value, into an iterable reader."""

    __slots__ = ("inner",)

    # noinspection PyProtocol
    inner: Callable[P, R]

    def __init__(self, inner: Callable[P, R]) -> None:  # noqa: D107
        self.inner = inner

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Iterator[R]:
        yield self.inner(*args, **kwargs)



class ProtobufValueError(ValueError):
    """Base class for value errors."""


class ProtobufTypeError(TypeError):
    """Base class for type errors."""


class IncorrectWireTypeError(ProtobufValueError):
    """Incorrect wire type in the stream."""


class UnexpectedWireTypeError(ProtobufValueError):
    """The wire type is correct, but a different one was expected."""


class IncorrectAnnotationError(ProtobufTypeError):
    """Something wrong with the attribute annotation."""


class UnsupportedAnnotationError(ProtobufTypeError):
    """The type annotation is not supported."""


class IncorrectValueError(ProtobufValueError):
    """Something's wrong with the field value."""



from abc import abstractmethod
from typing import IO, Protocol

from pure_protobuf.interfaces._repr import Repr


class Skip(Repr, Protocol):
    """
    Knows how to skip a value. It's equivalent to `Read`, but is supposed to be a little
    more efficient because it doesn't need to actually parse the field data.
    """

    @abstractmethod
    def __call__(self, io: IO[bytes]) -> None:
        raise NotImplementedError


class SkipNoOperation(Skip):
    def __call__(self, io: IO[bytes]) -> None:
        """Do nothing. This is useful for the deprecated groups feature."""


skip_no_operation = SkipNoOperation()



from io import SEEK_CUR
from typing import IO

from pure_protobuf.interfaces._skip import Skip


class SkipFixed32(Skip):
    """Skip fixed 32-field record."""

    def __call__(self, io: IO[bytes]) -> None:
        io.seek(4, SEEK_CUR)


skip_fixed_32 = SkipFixed32()



from typing import Protocol

from typing_extensions import get_args as get_type_args


class Repr(Protocol):
    """
    Provides default implementations for `str(self)` and `repr(self)` to make
    debugging easier.
    """

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        try:
            orig_class = self.__orig_class__  # type: ignore[attr-defined]
        except AttributeError:
            args = ""
        else:
            args = f"[{get_type_args(orig_class)[0]!r}]"
        return f"{type(self).__name__}{args}()"


class ReprWithInner(Repr, Protocol):
    inner: Repr

    def __repr__(self) -> str:
        try:
            orig_class = self.__orig_class__  # type: ignore[attr-defined]
        except AttributeError:
            args = ""
        else:
            args = f"[{get_type_args(orig_class)[0]!r}]"
        return f"{type(self).__name__}{args}({self.inner!r})"



"""Reading and writing parsed URLs."""

from typing import IO, Iterator
from urllib.parse import ParseResult, urlparse, urlunparse

from pure_protobuf.interfaces.read import Read
from pure_protobuf.interfaces.write import Write
from pure_protobuf.io.bytes_ import read_string, write_string


class ReadUrl(Read[ParseResult]):
    __slots__ = ()

    def __call__(self, io: IO[bytes]) -> Iterator[ParseResult]:
        yield urlparse(read_string(io))


class WriteUrl(Write[ParseResult]):
    __slots__ = ()

    def __call__(self, value: ParseResult, io: IO[bytes]) -> None:
        write_string(urlunparse(value), io)



from abc import abstractmethod
from typing import IO, Iterator, Protocol

from pure_protobuf.interfaces._repr import Repr
from pure_protobuf.interfaces._vars import RecordT_co
from pure_protobuf.io.wire_type import WireType


class ReadSingular(Repr, Protocol[RecordT_co]):
    @abstractmethod
    def __call__(self, __io: IO[bytes]) -> RecordT_co:
        raise NotImplementedError


class Read(Repr, Protocol[RecordT_co]):
    """Deserializes a value from the stream and yields it as an iterator."""

    @abstractmethod
    def __call__(self, __io: IO[bytes]) -> Iterator[RecordT_co]:
        raise NotImplementedError


class ReadTyped(Repr, Protocol[RecordT_co]):
    @abstractmethod
    def __call__(self, __io: IO[bytes], __actual_wire_type: WireType) -> Iterator[RecordT_co]:
        raise NotImplementedError



from collections.abc import Iterable
from typing import Any, List, Tuple, Union

from typing_extensions import TypeGuard, get_args, get_origin

try:
    from types import NoneType  # type: ignore
except ImportError:
    NoneType = type(None)  # type: ignore


class Sentinel:  # pragma: no cover
    """Sentinel object used for defaults."""

    def __str__(self) -> str:  # pragma: no cover
        return "DEFAULT"

    def __repr__(self) -> str:  # pragma: no cover
        return "DEFAULT"


DEFAULT = Sentinel()


def extract_repeated(hint: Any) -> Tuple[Any, TypeGuard[List]]:
    """Extract a possible repeated flag."""
    origin = get_origin(hint)
    if isinstance(origin, type) and (origin is Iterable or issubclass(origin, list)):
        return get_args(hint)[0], True
    return hint, False


def extract_optional(hint: Any) -> Tuple[Any, bool]:
    """Extract a possible optional flag."""
    if get_origin(hint) is Union:
        cleaned_args = tuple(arg for arg in get_args(hint) if arg is not NoneType)
        return Union[cleaned_args], True
    return hint, False



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Optional, Type, cast

from typing_extensions import Annotated
from typing_extensions import get_args as get_type_args
from typing_extensions import get_origin as get_type_origin

from pure_protobuf._accumulators import AccumulateAppend
from pure_protobuf._mergers import MergeConcatenate
from pure_protobuf.annotations import Field
from pure_protobuf.descriptors.record import RecordDescriptor
from pure_protobuf.helpers._dataclasses import KW_ONLY, SLOTS
from pure_protobuf.helpers._typing import extract_optional, extract_repeated
from pure_protobuf.interfaces._vars import FieldT, RecordT
from pure_protobuf.interfaces.accumulate import Accumulate
from pure_protobuf.interfaces.merge import Merge
from pure_protobuf.interfaces.read import ReadTyped
from pure_protobuf.interfaces.write import Write
from pure_protobuf.io.tag import Tag
from pure_protobuf.io.wire_type import WireType
from pure_protobuf.io.wrappers import (
    WriteLengthDelimited,
    WriteOptional,
    WriteRepeated,
    WriteTagged,
)

if TYPE_CHECKING:
    from pure_protobuf.annotations import OneOf
    from pure_protobuf.message import BaseMessage


@dataclass(frozen=True, **SLOTS, **KW_ONLY)
class _FieldDescriptor(Generic[FieldT, RecordT]):
    """
    Describes how the field should be read and written.

    Notes:
        - It's not a «real» Python descriptor, just didn't come up with a better name.
        - It gets converted from `RecordDescriptor`.
    """

    number: int
    """
    Field's number.

    See Also:
        - https://developers.google.com/protocol-buffers/docs/proto3#assigning_field_numbers
    """

    one_of: Optional[OneOf]
    """
    See Also:
        - https://developers.google.com/protocol-buffers/docs/proto3#oneof
    """

    write: Write[FieldT]
    read: ReadTyped[RecordT]
    accumulate: Accumulate[FieldT, RecordT]
    merge: Merge[FieldT]

    @classmethod
    def from_attribute(
        cls,
        message_type: Type[BaseMessage],
        attribute_hint: Any,
    ) -> Optional[_FieldDescriptor[Any, Any]]:
        """
        Construct a descriptor from the attribute's type hint.

        Returns:
            Field's descriptor, or `None` if skipped.
        """
        if get_type_origin(attribute_hint) is not Annotated:
            # It's not an annotated attribute, so ignore it.
            return None

        inner_hint, *annotated_args = get_type_args(attribute_hint)

        # noinspection PyProtectedMember
        field = Field._from_annotated_args(*annotated_args)
        if field is None:
            # `Field` annotation is missing, ignore the attribute.
            return None

        # Extract the flags from the hint.
        inner_hint, _ = extract_optional(inner_hint)
        inner_hint, is_repeated = extract_repeated(inner_hint)

        inner: RecordDescriptor[RecordT] = RecordDescriptor._from_inner_type_hint(
            message_type,
            inner_hint,
        )
        write = cast(Write[FieldT], inner.write)
        accumulate = cast(Accumulate[FieldT, RecordT], inner.accumulate)
        merge = cast(Merge[FieldT], inner.merge)

        # Abandon hope all ye who enter here.
        if field._packed_or(is_repeated and inner.wire_type.is_primitive_numeric):
            # Repeated fields of primitive numeric types are packed by default per the specification.
            if is_repeated:
                # Repeated packed field are untagged internally.
                write = WriteRepeated(write)

            # Now we can wrap it into a length-delimited value.
            write = WriteLengthDelimited(write)

            # And finally, tag it. The wire type is overridden.
            tag = Tag(field_number=field.number, wire_type=WireType.LEN)
            write = WriteTagged(write, tag)
        else:
            # In unpacked fields each value is separately tagged, thus applied first.
            tag = Tag(field_number=field.number, wire_type=inner.wire_type)
            write = WriteTagged(write, tag)
            if is_repeated:
                # `read` still reads one record at a time.
                write = WriteRepeated(write)

        if is_repeated:
            # Merger should concatenate repeated fields.
            merge = cast(Merge[FieldT], MergeConcatenate[RecordT]())
            accumulate = cast(Accumulate[FieldT, RecordT], AccumulateAppend[RecordT]())

        # And now just build and return the final descriptor.
        return cls(
            number=field.number,
            one_of=field.one_of,
            write=WriteOptional(write),
            read=inner.read,
            accumulate=accumulate,
            merge=merge,
        )



from abc import abstractmethod
from typing import IO, Protocol

from pure_protobuf.interfaces._repr import Repr
from pure_protobuf.interfaces._vars import FieldT_contra


class Write(Repr, Protocol[FieldT_contra]):
    """Serializes and writes the value into the stream."""

    @abstractmethod
    def __call__(self, __value: FieldT_contra, __io: IO[bytes]) -> None:
        raise NotImplementedError



from typing import Generic, Iterable, List, Optional, Type

from pure_protobuf.interfaces._repr import ReprWithInner
from pure_protobuf.interfaces._vars import MessageT, RecordT
from pure_protobuf.interfaces.accumulate import Accumulate


class AccumulateLastOneWins(Accumulate[RecordT, RecordT], Generic[RecordT]):
    """
    The simplest accumulator which keeps only the last seen item.

    Typical use case is fields of primitive scalar types.

    See Also:
        - https://developers.google.com/protocol-buffers/docs/encoding#last-one-wins
    """

    def __call__(self, accumulator: Optional[RecordT], other: Iterable[RecordT]) -> RecordT:
        for accumulator in other:  # noqa: B007
            pass
        return accumulator


class AccumulateAppend(Accumulate[List[RecordT], RecordT]):
    def __call__(
        self,
        accumulator: Optional[List[RecordT]],
        other: Iterable[RecordT],
    ) -> List[RecordT]:
        """Append all items from the `other` into the accumulator."""
        if accumulator is None:
            accumulator = []
        accumulator.extend(other)
        return accumulator


class AccumulateMessages(Accumulate[MessageT, MessageT], ReprWithInner):
    inner: Type[MessageT]

    __slots__ = ("inner",)

    # noinspection PyProtocol
    def __init__(self, inner: Type[MessageT]) -> None:
        self.inner = inner

    def __call__(self, lhs: Optional[MessageT], rhs: Iterable[MessageT]) -> MessageT:
        """
        Merge the two messages into the left one.

        Notes:
            - **Never** reuse `lhs` or `rhs` afterwards, consider them consumed by the method.
        """
        for other in rhs:
            assert other is not None, "right-hand side items must not be `None`"
            if lhs is not None:
                for name, descriptor in other.__PROTOBUF_FIELDS_BY_NUMBER__.values():
                    setattr(
                        lhs,
                        name,
                        descriptor.merge(getattr(lhs, name), getattr(other, name)),
                    )
                    one_of = descriptor.one_of
                    if one_of is not None:
                        one_of._keep_attribute(lhs, descriptor.number)
            else:
                lhs = other

        assert lhs is not None, "there must be at least one right-hand side item"
        return lhs



"""Well-known types."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from io import BytesIO
from typing import Any, Mapping, Optional, Type, cast
from urllib.parse import ParseResult

from typing_extensions import Annotated

from pure_protobuf.annotations import Field
from pure_protobuf.helpers._dataclasses import KW_ONLY, SLOTS
from pure_protobuf.helpers.datetime import split_seconds, unsplit_seconds
from pure_protobuf.message import BaseMessage


@dataclass(**KW_ONLY, **SLOTS)
class _TimeSpan(BaseMessage):
    """Base class to represent timespan as whole seconds plus its nanoseconds part."""

    seconds: Annotated[int, Field(1)] = 0
    nanos: Annotated[int, Field(2)] = 0


@dataclass(**KW_ONLY, **SLOTS)
class Timestamp(_TimeSpan):
    """
    Implements the [`Timestamp`](https://protobuf.dev/reference/protobuf/google.protobuf/#timestamp)
    well-known type and supports conversion from and to
    [`datetime`](https://docs.python.org/3/library/datetime.html#datetime-objects).
    """

    @classmethod
    def from_datetime(cls, value: datetime) -> Timestamp:
        """Convert the `datetime` to `Timestamp`."""
        seconds, nanos = split_seconds(value.timestamp())
        return cls(seconds=seconds, nanos=nanos)

    def into_datetime(self) -> datetime:
        """Convert to `datetime`."""
        return datetime.fromtimestamp(unsplit_seconds(self.seconds, self.nanos), tz=timezone.utc)


@dataclass(**KW_ONLY, **SLOTS)
class Duration(_TimeSpan):
    """
    Implements the [`#!protobuf Duration`](https://developers.google.com/protocol-buffers/docs/reference/google.protobuf#duration)
    well-known type.
    """

    @classmethod
    def from_timedelta(cls, value: timedelta) -> Duration:
        """Convert the `timedelta` into `Duration`."""
        seconds, nanos = split_seconds(value.total_seconds())
        return cls(seconds=seconds, nanos=nanos)

    def into_timedelta(self) -> timedelta:
        """Convert into `timedelta`."""
        return timedelta(seconds=unsplit_seconds(self.seconds, self.nanos))


@dataclass(**KW_ONLY, **SLOTS)
class Any_:  # noqa: N801
    """
    Well-known `Any` type.

    See Also:
        - https://github.com/protocolbuffers/protobuf/blob/master/src/google/protobuf/any.proto.
    """

    type_url: Annotated[ParseResult, Field(1)]
    value: Annotated[bytes, Field(2)] = b""

    @classmethod
    def from_message(cls, message: BaseMessage) -> Any_:
        """Convert the message into its `Any_` representation."""

        # noinspection PyArgumentList
        return cls(
            type_url=ParseResult(
                scheme="import",
                netloc=message.__module__,
                path=type(message).__qualname__,
                params="",
                query="",
                fragment="",
            ),
            value=bytes(message),
        )

    def into_message(
        self,
        locals_: Optional[Mapping[str, Any]] = None,
        globals_: Optional[Mapping[str, Any]] = None,
    ) -> BaseMessage:
        """
        Reconstructs a message from the current `Any_` representation.

        Args:
            locals_: forwarded to the `__import__` call
            globals_: forwarded to the `__import__` call

        See Also:
            - https://developers.google.com/protocol-buffers/docs/proto3#any
        """

        module = __import__(
            self.type_url.netloc,
            fromlist=[self.type_url.path],
            locals=locals_,
            globals=globals_,
        )
        class_ = cast(Type[BaseMessage], getattr(module, self.type_url.path))
        return class_.read_from(BytesIO(self.value))



"""Backwards compatibility for Python 3.8 and 3.9."""

from sys import version_info

if version_info >= (3, 10):
    KW_ONLY = {"kw_only": True}
    SLOTS = {"slots": True}
else:
    KW_ONLY = {}
    SLOTS = {}



"""Field annotations that may be used inside `Annotated`."""

from __future__ import annotations

from dataclasses import dataclass
from sys import version_info
from typing import TYPE_CHECKING, Any, ClassVar, NewType, Optional, Union

from pure_protobuf.exceptions import IncorrectAnnotationError
from pure_protobuf.helpers._dataclasses import SLOTS
from pure_protobuf.helpers._typing import DEFAULT, Sentinel

if version_info >= (3, 10):
    from dataclasses import KW_ONLY  # type: ignore[attr-defined]

if TYPE_CHECKING:
    from pure_protobuf.one_of import OneOf


@dataclass(frozen=True, **SLOTS)
class Field:
    """Annotates a Protocol Buffers field."""

    number: int
    """
    Specifies the field's number.

    See also:
        - https://developers.google.com/protocol-buffers/docs/proto3#assigning_field_numbers
    """

    if version_info >= (3, 10):
        _: ClassVar[KW_ONLY]

    packed: Union[bool, Sentinel] = DEFAULT
    """Specifies whether the field should be packed in its serialized representation."""

    one_of: Optional[OneOf] = None
    """Specifies a one-of group for this field."""

    @classmethod
    def _from_annotated_args(cls, *args: Any) -> Optional[Field]:
        """Extract itself from the `Annotated[_, *args]` type hint, if present."""
        for arg in args:
            if isinstance(arg, Field):
                arg._validate()
                return arg
        return None

    def _validate(self) -> None:
        if not 1 <= self.number <= 536_870_911:
            raise IncorrectAnnotationError(
                f"field number {self.number} is outside the allowed range",
            )
        if 19000 <= self.number <= 19999:
            raise IncorrectAnnotationError(f"field number {self.number} is reserved")

    def _packed_or(self, default: bool) -> bool:
        return self.packed if isinstance(self.packed, bool) else default


double = NewType("double", float)
"""
Fixed 64-bit floating point.

See Also:
    - https://developers.google.com/protocol-buffers/docs/proto3#scalar
    - https://developers.google.com/protocol-buffers/docs/encoding#non-varint-nums
"""

fixed32 = NewType("fixed32", int)
"""
Fixed unsigned 32-bit integer.

See Also:
    - https://developers.google.com/protocol-buffers/docs/encoding#non-varint-nums
"""

fixed64 = NewType("fixed64", int)
"""
Fixed unsigned 64-bit integer.

See Also:
    - https://developers.google.com/protocol-buffers/docs/encoding#non-varint-nums
"""

sfixed32 = NewType("sfixed32", int)
"""
Fixed signed 32-bit integer.

See Also:
    - https://developers.google.com/protocol-buffers/docs/encoding#non-varint-nums
"""

sfixed64 = NewType("sfixed64", int)
"""
Fixed signed 64-bit integer.

See Also:
    - https://developers.google.com/protocol-buffers/docs/proto3#scalar
    - https://developers.google.com/protocol-buffers/docs/encoding#non-varint-nums
"""

uint = NewType("uint", int)
"""
Unsigned variable-length integer.
"""

ZigZagInt = NewType("ZigZagInt", int)
"""
ZigZag-encoded variable-length integer.

See Also:
    - https://developers.google.com/protocol-buffers/docs/encoding#signed-ints
"""



from enum import IntEnum


class WireType(IntEnum):
    """
    See Also:
        - https://developers.google.com/protocol-buffers/docs/encoding#structure.
    """

    VARINT = 0
    """
    Base 128 Varints: `int32`, `int64`, `uint32`, `uint64`, `sint32`, `sint64`, `bool`, `enum`.

    See Also:
        - https://developers.google.com/protocol-buffers/docs/encoding#varints.
    """

    I64 = 1
    """Fixed 64-bit value: `fixed64`, `sfixed64`, `double`."""

    LEN = 2
    """String, bytes, embedded messages, packed repeated fields."""

    SGROUP = 3
    """
    Group start (deprecated).

    See Also:
        - https://developers.google.com/protocol-buffers/docs/encoding#groups
    """

    EGROUP = 4
    """
    Group end (deprecated).

    See Also:
        - https://developers.google.com/protocol-buffers/docs/encoding#groups
    """

    I32 = 5
    """Fixed 32-bit value: `fixed32`, `sfixed32`, `float`."""

    __PRIMITIVE_NUMERIC_TYPES__ = {VARINT, I32, I64}
    """
    See Also:
        - https://developers.google.com/protocol-buffers/docs/encoding#packed
    """

    @property
    def is_primitive_numeric(self) -> bool:
        return self in self.__PRIMITIVE_NUMERIC_TYPES__



from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from pure_protobuf.message import BaseMessage  # noqa

FieldT = TypeVar("FieldT")
"""Message field's type."""

FieldT_contra = TypeVar("FieldT_contra", contravariant=True)

RecordT = TypeVar("RecordT")
"""Field's record type. May be the field's type itself, or a single item out of it."""

RecordT_contra = TypeVar("RecordT_contra", contravariant=True)
RecordT_co = TypeVar("RecordT_co", covariant=True)

MessageT = TypeVar("MessageT", bound="BaseMessage")



from io import SEEK_CUR
from typing import IO

from pure_protobuf.interfaces._skip import Skip


class SkipFixed64(Skip):
    __slots__ = ()

    def __call__(self, io: IO[bytes]) -> None:
        io.seek(8, SEEK_CUR)


skip_fixed_64 = SkipFixed64()



from io import BytesIO
from typing import IO, Generic, Iterable, Iterator, Optional, cast

from pure_protobuf.exceptions import UnexpectedWireTypeError
from pure_protobuf.interfaces._repr import ReprWithInner
from pure_protobuf.interfaces._vars import FieldT_contra, RecordT
from pure_protobuf.interfaces.read import Read, ReadTyped
from pure_protobuf.interfaces.write import Write
from pure_protobuf.io.bytes_ import read_bytes, write_bytes
from pure_protobuf.io.tag import Tag
from pure_protobuf.io.wire_type import WireType


class ReadStrictlyTyped(ReadTyped[RecordT], ReprWithInner):
    """Verifies the actual wire type."""

    __slots__ = ("inner", "expected_wire_type")

    # noinspection PyProtocol
    def __init__(self, inner: Read[RecordT], expected_wire_type: WireType) -> None:
        self.inner = inner
        self.expected_wire_type = expected_wire_type

    def __call__(self, io: IO[bytes], actual_wire_type: WireType) -> Iterator[RecordT]:
        if actual_wire_type != self.expected_wire_type:
            raise UnexpectedWireTypeError(
                f"expected {self.expected_wire_type!r} but received {actual_wire_type!r}",
            )
        yield from self.inner(io)


class ReadMaybePacked(ReadTyped[RecordT], ReprWithInner):
    """Can be called for either packed or unpacked record."""

    __slots__ = ("inner", "inner_packed", "unpacked_wire_type")

    # noinspection PyProtocol
    def __init__(self, inner: Read[RecordT], unpacked_wire_type: WireType) -> None:
        self.inner = inner
        self.unpacked_wire_type = unpacked_wire_type
        self.inner_packed = ReadLengthDelimited[RecordT](ReadRepeated[RecordT](inner))

    def __call__(self, io: IO[bytes], actual_wire_type: WireType) -> Iterator[RecordT]:
        if actual_wire_type == self.unpacked_wire_type:
            yield from self.inner(io)
        elif actual_wire_type == WireType.LEN:
            yield from self.inner_packed(io)
        else:
            raise UnexpectedWireTypeError(
                f"expected {self.unpacked_wire_type!r} or a packed record but received {actual_wire_type!r}",
            )


class ReadLengthDelimited(Read[RecordT], ReprWithInner):
    """
    Make the inner reader length-delimited.

    This reader wraps the inner reader so that it would first read a byte string,
    and then call the inner reader on that byte string.
    """

    __slots__ = ("inner",)

    # noinspection PyProtocol
    def __init__(self, inner: Read[RecordT]) -> None:
        self.inner = inner

    def __call__(self, io: IO[bytes]) -> Iterator[RecordT]:
        yield from self.inner(BytesIO(read_bytes(io)))


class ReadRepeated(Read[RecordT], ReprWithInner):
    """Wrap an inner reader to read repeated records."""

    __slots__ = ("inner",)

    # noinspection PyProtocol
    def __init__(self, inner: Read[RecordT]) -> None:
        self.inner = inner

    def __call__(self, io: IO[bytes]) -> Iterator[RecordT]:
        while True:
            try:
                yield from self.inner(io)
            except EOFError:
                break


class WriteRepeated(Generic[RecordT, FieldT_contra], Write[FieldT_contra], ReprWithInner):
    """Wrap an inner writer to produce repeated records."""

    __slots__ = ("inner",)

    inner: Write[RecordT]

    # noinspection PyProtocol
    def __init__(self, inner: Write[RecordT]) -> None:
        self.inner = inner

    def __call__(self, values: FieldT_contra, io: IO[bytes]) -> None:
        for value in cast(Iterable[RecordT], values):
            self.inner(value, io)


def to_bytes(write: Write[RecordT], value: RecordT) -> bytes:
    io = BytesIO()
    write(value, io)
    return io.getvalue()


class WriteTagged(Write[RecordT], ReprWithInner):
    """Prepends the tag to a value being written."""

    __slots__ = ("inner", "encoded_tag")

    # noinspection PyProtocol
    def __init__(self, inner: Write[RecordT], tag: Tag) -> None:
        self.inner = inner
        self.encoded_tag = to_bytes(Tag.write_to, tag)

    def __call__(self, value: RecordT, io: IO[bytes]) -> None:
        io.write(self.encoded_tag)
        self.inner(value, io)


class WriteOptional(Write[RecordT], ReprWithInner):
    """Wrap an inner writer to skip serialization of `None`."""

    __slots__ = ("inner",)

    inner: Write[RecordT]

    # noinspection PyProtocol
    def __init__(self, inner: Write[RecordT]) -> None:
        self.inner = inner

    def __call__(self, value: Optional[RecordT], io: IO[bytes]) -> None:
        if value is not None:
            self.inner(value, io)


class WriteLengthDelimited(Write[RecordT], ReprWithInner):
    """Wrap an inner writer into a length-delimited record."""

    __slots__ = ("inner",)

    inner: Write[RecordT]

    # noinspection PyProtocol
    def __init__(self, inner: Write[RecordT]) -> None:
        self.inner = inner

    def __call__(self, value: RecordT, io: IO[bytes]) -> None:
        write_bytes(to_bytes(self.inner, value), io)



"""Input/output helpers."""

from typing import IO


def read_byte_checked(io: IO[bytes]) -> int:
    """
    Read one byte.

    Raises:
        EOFError: the stream has ended
    """
    buffer = io.read(1)
    if not buffer:
        raise EOFError
    return buffer[0]


def read_checked(io: IO[bytes], count: int) -> bytes:
    """
    Read the specified number of bytes.

    Raises:
        EOFError: the stream has ended
    """
    buffer = io.read(count)
    if len(buffer) < count:
        raise EOFError(f"{count} bytes expected, but only {len(buffer)} read")
    return buffer



"""
Reading and writing varints and the derivative types.

See Also:
    - https://developers.google.com/protocol-buffers/docs/encoding#varints

"""

from enum import IntEnum
from itertools import count
from sys import byteorder
from typing import IO, Iterator, Type, TypeVar

from pure_protobuf.exceptions import IncorrectValueError
from pure_protobuf.helpers.io import read_byte_checked
from pure_protobuf.interfaces._skip import Skip
from pure_protobuf.interfaces.read import Read, ReadSingular
from pure_protobuf.interfaces.write import Write


class SkipVarint(Skip):
    def __call__(self, io: IO[bytes]) -> None:
        while read_byte_checked(io) & 0x80:
            pass


class ReadUnsignedVarint(ReadSingular[int]):
    """Reads unsigned varint from the stream."""

    def __call__(self, io: IO[bytes]) -> int:
        value = 0
        for shift in count(0, 7):
            byte = read_byte_checked(io)
            value |= (byte & 0x7F) << shift
            if not byte & 0x80:
                return value
        raise AssertionError("unreachable code")


class WriteUnsignedVarint(Write[int]):
    """Writes unsigned varint to the stream."""

    def __call__(self, value: int, io: IO[bytes]) -> None:
        while value > 0x7F:
            io.write(bytes((value & 0x7F | 0x80,)))
            value >>= 7
        io.write(bytes((value,)))


skip_varint = SkipVarint()
read_unsigned_varint = ReadUnsignedVarint()
write_unsigned_varint = WriteUnsignedVarint()


class ReadZigZagVarint(ReadSingular[int]):
    """
    Reads a ZigZag-encoded varint.

    See Also:
        - https://stackoverflow.com/a/2211086/359730.
    """

    def __call__(self, io: IO[bytes]) -> int:
        value = read_unsigned_varint(io)
        return (value >> 1) ^ (-(value & 1))


class WriteZigZagVarint(Write[int]):
    """Writes a ZigZag-encoded varint."""

    def __call__(self, value: int, io: IO[bytes]) -> None:
        write_unsigned_varint(abs(value) * 2 - (value < 0), io)


class ReadTwosComplimentVarint(ReadSingular[int]):
    """
    Reads a two's compliment varint.

    See Also:
        - https://protobuf.dev/programming-guides/encoding/#signed-ints
    """

    def __call__(self, io: IO[bytes]) -> int:
        varint = read_unsigned_varint(io)
        return int.from_bytes(
            varint.to_bytes(8, byteorder, signed=False),
            byteorder,
            signed=True,
        )


class WriteTwosComplimentVarint(Write[int]):
    """
    Writes a two's compliment varint.

    See Also:
        - https://protobuf.dev/programming-guides/encoding/#signed-ints
    """

    def __call__(self, value: int, io: IO[bytes]) -> None:
        compliment = int.from_bytes(
            value.to_bytes(8, byteorder, signed=True),
            byteorder,
            signed=False,
        )
        return write_unsigned_varint(compliment, io)


class ReadBool(ReadSingular[bool]):
    def __call__(self, io: IO[bytes]) -> bool:
        return bool(read_unsigned_varint(io))


class WriteBool(Write[bool]):
    def __call__(self, value: bool, io: IO[bytes]) -> None:
        write_unsigned_varint(int(value), io)


read_bool = ReadBool()
write_bool = WriteBool()


EnumT = TypeVar("EnumT", bound=IntEnum)


class ReadEnum(Read[EnumT]):
    __slots__ = ("enum_type",)

    # noinspection PyProtocol
    def __init__(self, enum_type: Type[EnumT]) -> None:
        self.enum_type = enum_type

    def __call__(self, io: IO[bytes]) -> Iterator[EnumT]:
        value = read_unsigned_varint(io)
        try:
            yield self.enum_type(value)
        except ValueError as e:
            raise IncorrectValueError(
                f"incorrect value {value} for enum `{self.enum_type!r}`",
            ) from e

    def __repr__(self) -> str:  # noqa: D105
        return f"{type(self).__name__}({self.enum_type.__name__})"


class WriteEnum(Write[EnumT]):
    __slots__ = ()

    def __call__(self, value: EnumT, io: IO[bytes]) -> None:
        # noinspection PyTypeChecker
        write_unsigned_varint(value.value, io)



from __future__ import annotations

from abc import ABC
from io import BytesIO
from typing import IO, Any, ClassVar, Dict, Tuple

from typing_extensions import Self

try:
    from inspect import get_annotations  # type: ignore[attr-defined]
except ImportError:
    from get_annotations import get_annotations  # type: ignore[no-redef]

from pure_protobuf._accumulators import AccumulateMessages
from pure_protobuf._mergers import MergeMessages
from pure_protobuf.descriptors._field import _FieldDescriptor
from pure_protobuf.descriptors.record import RecordDescriptor
from pure_protobuf.helpers.itertools import ReadCallback
from pure_protobuf.interfaces._skip import Skip, skip_no_operation
from pure_protobuf.io.bytes_ import skip_bytes
from pure_protobuf.io.fixed32 import skip_fixed_32
from pure_protobuf.io.fixed64 import skip_fixed_64
from pure_protobuf.io.tag import Tag
from pure_protobuf.io.varint import skip_varint
from pure_protobuf.io.wire_type import WireType
from pure_protobuf.io.wrappers import (
    ReadLengthDelimited,
    ReadStrictlyTyped,
    WriteLengthDelimited,
    to_bytes,
)


class BaseMessage(ABC):
    """Base message class, inherit from it to define a specific message."""

    __slots__ = ()

    __PROTOBUF_FIELDS_BY_NUMBER__: ClassVar[Dict[int, Tuple[str, _FieldDescriptor]]]
    __PROTOBUF_FIELDS_BY_NAME__: ClassVar[Dict[str, _FieldDescriptor]]

    __PROTOBUF_SKIP__: ClassVar[Dict[WireType, Skip]] = {
        WireType.VARINT: skip_varint,
        WireType.I64: skip_fixed_64,
        WireType.LEN: skip_bytes,
        WireType.I32: skip_fixed_32,
        WireType.SGROUP: skip_no_operation,
        WireType.EGROUP: skip_no_operation,
    }
    """Defines how to skip a field of the given wire type."""

    def __init_subclass__(cls) -> None:  # noqa: D105
        cls.__PROTOBUF_FIELDS_BY_NUMBER__ = {}
        cls.__PROTOBUF_FIELDS_BY_NAME__ = {}

        type_hints: Dict[str, Any] = get_annotations(cls, eval_str=True)
        for name, hint in type_hints.items():
            descriptor = _FieldDescriptor.from_attribute(cls, hint)
            if descriptor is not None:
                cls.__PROTOBUF_FIELDS_BY_NUMBER__[descriptor.number] = (name, descriptor)
                cls.__PROTOBUF_FIELDS_BY_NAME__[name] = descriptor
                one_of = descriptor.one_of
                if one_of is not None:
                    one_of._add_field(descriptor.number, name)

    @classmethod
    def read_from(cls, io: IO[bytes]) -> Self:
        """Read a message from the file."""

        values: Dict[str, Any] = {}
        while True:
            try:
                tag = Tag.read_from(io)
            except EOFError:
                break
            try:
                name, descriptor = cls.__PROTOBUF_FIELDS_BY_NUMBER__[tag.field_number]
            except KeyError:
                # The field is not defined, skip it in the stream.
                # TODO: https://github.com/eigenein/protobuf/issues/99.
                cls.__PROTOBUF_SKIP__[tag.wire_type](io)
            else:
                # Read the value and accumulate it.
                values[name] = descriptor.accumulate(
                    values.get(name),
                    descriptor.read(io, tag.wire_type),
                )

                # Possibly update the one-of field.
                one_of = descriptor.one_of
                if one_of is not None:
                    one_of._keep_values(values, descriptor.number)

        return cls(**values)

    @classmethod
    def loads(cls, buffer: bytes) -> Self:
        """
        Read a message from the buffer.

        This is functionally the same as calling `read_from(BytesIO(buffer))`.
        """
        return cls.read_from(BytesIO(buffer))

    def write_to(self, io: IO[bytes]) -> None:
        """Write the message to the file."""
        for _, (name, descriptor) in self.__PROTOBUF_FIELDS_BY_NUMBER__.items():
            descriptor.write(getattr(self, name), io)

    def __bytes__(self) -> bytes:
        """
        Convert the message to a bytestring.

        This is functionally the same as calling `dumps()` or `write_to(BytesIO(…))`.
        """
        return to_bytes(BaseMessage.write_to, self)

    def dumps(self) -> bytes:
        """
        Convert the message to a bytestring.

        This is functionally the same as calling `bytes(message)` or `write_to(BytesIO(…))`.
        """
        return bytes(self)

    def __setattr__(self, name: str, value: Any) -> None:  # noqa: D105
        super().__setattr__(name, value)
        descriptor = self.__PROTOBUF_FIELDS_BY_NAME__[name]
        one_of = descriptor.one_of
        if one_of is not None:
            one_of._keep_attribute(self, descriptor.number)

    @classmethod
    def _init_embedded_descriptor(cls) -> RecordDescriptor[Self]:
        accumulate = AccumulateMessages(cls)
        return RecordDescriptor(
            wire_type=WireType.LEN,
            write=WriteLengthDelimited(cls.write_to),
            read=ReadStrictlyTyped(ReadLengthDelimited(ReadCallback(cls.read_from)), WireType.LEN),
            accumulate=accumulate,
            merge=MergeMessages(accumulate),
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING, Any, ByteString, ClassVar, Dict, Generic, Type
from urllib.parse import ParseResult

from typing_extensions import Self

from pure_protobuf._accumulators import AccumulateLastOneWins
from pure_protobuf._mergers import MergeLastOneWins
from pure_protobuf.annotations import ZigZagInt, double, fixed32, fixed64, sfixed32, sfixed64, uint
from pure_protobuf.exceptions import UnsupportedAnnotationError
from pure_protobuf.helpers._dataclasses import KW_ONLY, SLOTS
from pure_protobuf.helpers.itertools import ReadCallback
from pure_protobuf.interfaces._vars import RecordT
from pure_protobuf.interfaces.accumulate import Accumulate
from pure_protobuf.interfaces.merge import Merge
from pure_protobuf.interfaces.read import ReadTyped
from pure_protobuf.interfaces.write import Write
from pure_protobuf.io.bytes_ import read_bytes, read_string, write_bytes, write_string
from pure_protobuf.io.struct_ import ReadStruct, WriteStruct
from pure_protobuf.io.url import ReadUrl, WriteUrl
from pure_protobuf.io.varint import (
    ReadEnum,
    ReadTwosComplimentVarint,
    ReadZigZagVarint,
    WriteEnum,
    WriteTwosComplimentVarint,
    WriteZigZagVarint,
    read_bool,
    read_unsigned_varint,
    write_bool,
    write_unsigned_varint,
)
from pure_protobuf.io.wire_type import WireType
from pure_protobuf.io.wrappers import ReadMaybePacked, ReadStrictlyTyped

if TYPE_CHECKING:
    from pure_protobuf.message import BaseMessage


@dataclass(**KW_ONLY, **SLOTS)
class RecordDescriptor(Generic[RecordT]):
    """Describes how records should be read, written, accumulated, and merged. Not a «real» Python descriptor."""

    wire_type: WireType
    """Field's record [wire type](https://developers.google.com/protocol-buffers/docs/encoding#structure)."""

    write: Write[RecordT]
    """Write a complete value to the stream, altogether with its tag."""

    read: ReadTyped[RecordT]
    """
    Read a record from the stream.

    This behaves differently from the [`write`][pure_protobuf.descriptors.record.RecordDescriptor.write],
    because it's only supposed to read a single record from the stream
    (it may be, for example, just one item of a packed repeated field).
    Also, it assumes that the tag has already been read by [`BaseMessage`][base-message].
    """

    accumulate: Accumulate[RecordT, RecordT] = AccumulateLastOneWins()
    """
    Accumulate a value from the stream into an existing field value.
    It follows the `read` to decide which value should be assigned to the attribute.
    """

    merge: Merge[RecordT] = MergeLastOneWins()
    """Merge two values of the same field from different messages. Only called in a message merger."""

    __PREDEFINED__: ClassVar[Dict[Any, RecordDescriptor]]
    """Pre-defined descriptors for primitive types."""

    @classmethod
    def _from_inner_type_hint(
        cls,
        message_type: Type[BaseMessage],
        inner_hint: Any,
    ) -> RecordDescriptor[Any]:
        """
        Construct a descriptor from the inner type hint.

        Examples:
            - For `Annotated[int, Field[1]]` the inner hint is `int`.
            - For `Annotated[List[int], Field[1]]` it's also `int`
              since the `List` has already been extracted by `Field._from_annotated_args`.

        Args:
            message_type: message type which contains the attribute being described
            inner_hint: the attribute's own type hint
        """

        from pure_protobuf.message import BaseMessage

        try:
            singular = RecordDescriptor.__PREDEFINED__[inner_hint]
        except KeyError:
            pass
        else:
            return RecordDescriptor(
                wire_type=singular.wire_type,
                read=singular.read,
                write=singular.write,
            )

        if inner_hint is Self:
            # Support recursive types.
            return message_type._init_embedded_descriptor()

        if isinstance(inner_hint, type):
            if issubclass(inner_hint, IntEnum):
                return RecordDescriptor(
                    wire_type=WireType.VARINT,
                    write=WriteEnum[inner_hint](),
                    read=ReadMaybePacked[inner_hint](ReadEnum(inner_hint), WireType.VARINT),
                )
            if issubclass(inner_hint, BaseMessage):
                return inner_hint._init_embedded_descriptor()

        raise UnsupportedAnnotationError(f"type annotation `{inner_hint!r}` is not supported")


BOOL_DESCRIPTOR: RecordDescriptor[bool] = RecordDescriptor(
    wire_type=WireType.VARINT,
    write=write_bool,
    read=ReadMaybePacked[bool](ReadCallback(read_bool), WireType.VARINT),
)
BYTES_DESCRIPTOR: RecordDescriptor[bytes] = RecordDescriptor(
    wire_type=WireType.LEN,
    write=write_bytes,
    read=ReadStrictlyTyped(ReadCallback(read_bytes), WireType.LEN),
)
FLOAT_DESCRIPTOR: RecordDescriptor[float] = RecordDescriptor(
    wire_type=WireType.I32,
    read=ReadMaybePacked(ReadStruct[float]("<f"), WireType.I32),
    write=WriteStruct[float]("<f"),
)
DOUBLE_DESCRIPTOR: RecordDescriptor[double] = RecordDescriptor(
    wire_type=WireType.I64,
    read=ReadMaybePacked(ReadStruct[double]("<d"), WireType.I64),
    write=WriteStruct[double]("<d"),
)
SIGNED_INT32_DESCRIPTOR: RecordDescriptor[sfixed32] = RecordDescriptor(
    wire_type=WireType.I32,
    read=ReadMaybePacked(ReadStruct[sfixed32]("<i"), WireType.I32),
    write=WriteStruct[sfixed32]("<i"),
)
UNSIGNED_INT32_DESCRIPTOR: RecordDescriptor[fixed32] = RecordDescriptor(
    wire_type=WireType.I32,
    read=ReadMaybePacked(ReadStruct[fixed32]("<I"), WireType.I32),
    write=WriteStruct[fixed32]("<I"),
)
SIGNED_INT64_DESCRIPTOR: RecordDescriptor[sfixed64] = RecordDescriptor(
    wire_type=WireType.I64,
    read=ReadMaybePacked(ReadStruct[sfixed64]("<i"), WireType.I64),
    write=WriteStruct[sfixed64]("<q"),
)
UNSIGNED_INT64_DESCRIPTOR: RecordDescriptor[fixed64] = RecordDescriptor(
    wire_type=WireType.I64,
    read=ReadMaybePacked(ReadStruct[fixed64]("<I"), WireType.I64),
    write=WriteStruct[fixed64]("<Q"),
)
URL_DESCRIPTOR: RecordDescriptor[ParseResult] = RecordDescriptor(
    wire_type=WireType.LEN,
    read=ReadStrictlyTyped[ParseResult](ReadUrl(), WireType.LEN),
    write=WriteUrl(),
)

RecordDescriptor.__PREDEFINED__ = {
    bool: BOOL_DESCRIPTOR,
    bytes: BYTES_DESCRIPTOR,
    bytearray: BYTES_DESCRIPTOR,
    ByteString: BYTES_DESCRIPTOR,
    fixed32: UNSIGNED_INT32_DESCRIPTOR,
    fixed64: UNSIGNED_INT64_DESCRIPTOR,
    float: FLOAT_DESCRIPTOR,
    double: DOUBLE_DESCRIPTOR,
    int: RecordDescriptor(
        wire_type=WireType.VARINT,
        write=WriteTwosComplimentVarint(),
        read=ReadMaybePacked[int](ReadCallback(ReadTwosComplimentVarint()), WireType.VARINT),
    ),
    memoryview: BYTES_DESCRIPTOR,
    ParseResult: URL_DESCRIPTOR,
    sfixed32: SIGNED_INT32_DESCRIPTOR,
    sfixed64: UNSIGNED_INT64_DESCRIPTOR,
    str: RecordDescriptor(
        wire_type=WireType.LEN,
        write=write_string,
        read=ReadStrictlyTyped(ReadCallback(read_string), WireType.LEN),
    ),
    uint: RecordDescriptor(
        wire_type=WireType.VARINT,
        write=write_unsigned_varint,
        read=ReadMaybePacked[int](ReadCallback(read_unsigned_varint), WireType.VARINT),
    ),
    ZigZagInt: RecordDescriptor(
        wire_type=WireType.VARINT,
        write=WriteZigZagVarint(),
        read=ReadMaybePacked[int](ReadCallback(ReadZigZagVarint()), WireType.VARINT),
    ),
}



from __future__ import annotations

from typing import Any, Callable, Generic, List, MutableMapping, Optional, Tuple, Type, TypeVar

from pure_protobuf.message import BaseMessage

OneOfT = TypeVar("OneOfT")


class OneOf(Generic[OneOfT]):
    """
    See Also:
        - https://developers.google.com/protocol-buffers/docs/proto3#oneof.
    """

    __slots__ = ("_fields",)

    def __init__(self) -> None:
        self._fields: List[Tuple[int, str]] = []

    def _add_field(self, number: int, name: str) -> None:
        self._fields.append((number, name))

    def _keep_values(
        self,
        values: MutableMapping[str, Any],
        keep_number: int,
    ) -> None:
        for other_number, other_name in self._fields:
            if other_number != keep_number:
                values.pop(other_name, None)

    def _keep_attribute(self, message: BaseMessage, keep_number: int) -> None:
        for other_number, other_name in self._fields:
            if other_number != keep_number:
                super(BaseMessage, message).__setattr__(other_name, None)

    def __get__(self, instance: Any, type_: Type[Any]) -> Optional[OneOfT]:  # noqa: D105
        if not isinstance(instance, BaseMessage):
            # Allows passing the descriptor by reference, and we need to move the descriptor from
            # the corresponding annotation.
            # This is not a part of the public interface, hence the «type: ignore».
            return self  # type: ignore[return-value]
        name = self._which_one_of(instance)
        return getattr(instance, name) if name else None

    def __set__(self, instance: BaseMessage, _value: Any) -> None:  # noqa: D105
        raise RuntimeError("attempted to set the one-of field, use a specific attribute instead")

    def which_one_of_getter(self) -> Callable[[BaseMessage], Optional[str]]:
        """Construct a getter which returns which attribute is actually set."""

        def which_one_of(instance: BaseMessage) -> Optional[str]:
            return self._which_one_of(instance)

        return which_one_of

    def _which_one_of(self, instance: BaseMessage) -> Optional[str]:
        """Return which of the attributes is actually set."""
        for _, name in self._fields:
            value = getattr(instance, name)
            if value is not None:
                return name
        return None



from dataclasses import dataclass
from typing import IO

from typing_extensions import Self

from pure_protobuf.exceptions import IncorrectWireTypeError
from pure_protobuf.helpers._dataclasses import KW_ONLY, SLOTS
from pure_protobuf.io.varint import read_unsigned_varint, write_unsigned_varint
from pure_protobuf.io.wire_type import WireType


@dataclass(frozen=True, **KW_ONLY, **SLOTS)
class Tag:
    """
    Represents a field tag: a pair of key and wire type.

    See Also:
        - https://developers.google.com/protocol-buffers/docs/encoding#structure.
    """

    field_number: int
    wire_type: WireType

    @classmethod
    def read_from(cls, io: IO[bytes]) -> Self:
        """Read a tag from the file object."""
        encoded_tag = read_unsigned_varint(io)
        encoded_wire_type = encoded_tag & 0b111
        try:
            wire_type = WireType(encoded_wire_type)
        except ValueError as e:
            raise IncorrectWireTypeError(f"incorrect wire type {encoded_wire_type}") from e
        return cls(field_number=(encoded_tag >> 3), wire_type=wire_type)

    def encode(self) -> int:
        """Convert the tag into its encoded format."""
        return (self.field_number << 3) | self.wire_type.value  # type: ignore

    def write_to(self, io: IO[bytes]) -> None:
        """Write the tag into the file object."""
        write_unsigned_varint(self.encode(), io)



from abc import abstractmethod
from typing import Iterable, Optional, Protocol

from pure_protobuf.interfaces._repr import Repr
from pure_protobuf.interfaces._vars import FieldT, RecordT_contra


class Accumulate(Repr, Protocol[FieldT, RecordT_contra]):
    """
    Accumulates the value in the accumulator (also known as «fold» or «aggregate»).
    Implementation of the protocol defines how a record value from the input buffer
    should be «merged» into a corresponding attribute.
    """

    @abstractmethod
    def __call__(
        self,
        __accumulator: Optional[FieldT],
        __values: Iterable[RecordT_contra],
    ) -> FieldT:
        raise NotImplementedError(f"{type(self)}.__call__()")



"""
Reading and writing scalar length-delimited values.

See Also:
    - https://developers.google.com/protocol-buffers/docs/encoding#length-types

"""

from io import SEEK_CUR
from typing import IO

from pure_protobuf.helpers.io import read_checked
from pure_protobuf.interfaces._skip import Skip
from pure_protobuf.interfaces.read import ReadSingular
from pure_protobuf.interfaces.write import Write
from pure_protobuf.io.varint import read_unsigned_varint, write_unsigned_varint


class SkipBytes(Skip):
    def __call__(self, io: IO[bytes]) -> None:
        length = read_unsigned_varint(io)
        io.seek(length, SEEK_CUR)


class WriteBytes(Write[bytes]):
    def __call__(self, value: bytes, io: IO[bytes]) -> None:
        write_unsigned_varint(len(value), io)
        io.write(value)


class ReadBytes(ReadSingular[bytes]):
    def __call__(self, io: IO[bytes]) -> bytes:
        length = read_unsigned_varint(io)
        return read_checked(io, length)


skip_bytes = SkipBytes()
read_bytes = ReadBytes()
write_bytes = WriteBytes()


class ReadString(ReadSingular[str]):
    def __call__(self, io: IO[bytes]) -> str:
        return read_bytes(io).decode("utf-8")


class WriteString(Write[str]):
    def __call__(self, value: str, io: IO[bytes]) -> None:
        write_bytes(value.encode("utf-8"), io)


read_string = ReadString()
write_string = WriteString()



from struct import Struct
from typing import IO, Generic, Iterator

from pure_protobuf.helpers.io import read_checked
from pure_protobuf.interfaces._repr import ReprWithInner
from pure_protobuf.interfaces._vars import RecordT_co, RecordT_contra
from pure_protobuf.interfaces.read import Read
from pure_protobuf.interfaces.write import Write


class ReadStruct(Read[RecordT_co], ReprWithInner, Generic[RecordT_co]):
    inner: Struct

    __slots__ = ("inner",)

    # noinspection PyProtocol
    def __init__(self, format_: str) -> None:  # noqa: D107
        self.inner = Struct(format_)

    def __call__(self, io: IO[bytes]) -> Iterator[RecordT_co]:
        inner = self.inner
        yield from inner.unpack(read_checked(io, inner.size))


class WriteStruct(Write[RecordT_contra], ReprWithInner, Generic[RecordT_contra]):
    inner: Struct

    __slots__ = ("inner",)

    # noinspection PyProtocol
    def __init__(self, format_: str) -> None:  # noqa: D107
        self.inner = Struct(format_)

    def __call__(self, value: RecordT_contra, io: IO[bytes]) -> None:
        io.write(self.inner.pack(value))
